/*
edentity.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "edentity.h"

#include "3d.h"
#include "brush.h"
#include "button.h"
#include "color.h"
#include "edent.h"
#include "entity.h"
#include "error.h"
#include "file.h"
#include "keyboard.h"
#include "list.h"
#include "memory.h"
#include "message.h"
#include "mouse.h"
#include "popup.h"
#include "popupwin.h"
#include "entclass.h"
#include "quest.h"
#include "qui.h"
#include "tex.h"
#include "texpick.h"
#include "video.h"


static int EditCol(char *buf,int col_max,int col_scale)
{
   QUI_window_t *w;

   unsigned char *TempBuf;

   int bp;
   int b_ok,b_cancel,b_tex;
   int b_inc[3];
   int b_dec[3];
   int b_scale;

   float col[3];
   float max;

   int i;
   int update;


   SetPal(PAL_QUEST);

   col[0]=col[1]=col[2]=0;
   sscanf(buf,"%f %f %f",&col[0],&col[1],&col[2]);

   if (col_scale)
   {
      max=col[0];
      for (i=1;i<3;i++)
         if (col[i]>max)
            max=col[i];
   
      if (max)
      {
         max=col_max/max;
         for (i=0;i<3;i++)
            col[i]*=max;
      }
   }

	w = &Q.window[POP_WINDOW_1+Q.num_popups];

	w->size.x = 200;
	w->pos.x = (video.ScreenWidth / 2) - (w->size.x / 2);

	w->size.y = 192;
	w->pos.y = (video.ScreenHeight / 2) - (w->size.y / 2);

	QUI_PopUpWindow(POP_WINDOW_1+Q.num_popups,"Edit color",&TempBuf);
	Q.num_popups++;

   PushButtons();
   b_ok    =AddButtonText(0,0,B_ENTER,"OK");
   b_cancel=AddButtonText(0,0,B_ESCAPE,"Cancel");
   b_tex   =AddButtonText(0,0,0,"Texture");
   b_scale =AddButtonText(0,0,0,"Scale");

   button[b_ok].x=w->pos.x+4;
   button[b_cancel].x=button[b_ok].x+button[b_ok].sx+4;
   button[b_tex].x=button[b_cancel].x+button[b_cancel].sx+4;

   button[b_ok].y=button[b_cancel].y=button[b_tex].y=
      w->pos.y + w->size.y - button[b_ok].sy - 4;

   button[b_scale].x=button[b_ok].x;
   button[b_scale].y=button[b_ok].y-button[b_scale].sy-4;

   for (i=0;i<3;i++)
   {
      b_inc[i]=AddButtonPic(w->pos.x+8,w->pos.y+32+32*i,
         B_RAPID,"button_tiny_up");
      b_dec[i]=AddButtonPic(w->pos.x+8,w->pos.y+32+32*i+16,
         B_RAPID,"button_tiny_down");
   }

   DrawButtons();

   for (i=0;i<32;i++)
      DrawLine(w->pos.x+128   ,w->pos.y+48+i,
               w->pos.x+128+32,w->pos.y+48+i,255);

	RefreshPart(w->pos.x,w->pos.y,w->pos.x+w->size.x,w->pos.y+w->size.y);

   update=1;
   while (1)
   {
      if (update)
      {
         for (i=0;i<3;i++)
         {
            col[i]=(int)(col[i]);
            if (col[i]<0) col[i]=0;
            if (col[i]>col_max) col[i]=col_max;

            QUI_DrawStr(w->pos.x+24,w->pos.y+32+32*i+4,BG_COLOR,0,0,0,
               "%4.0f  ",col[i]);
         }

         for (i=0;i<3;i++)
            video.pal[255*3+i]=col[i]*(64.0/(float)col_max);
         SetGammaPal(video.pal);

         RefreshPart(w->pos.x,w->pos.y,
                     w->pos.x+w->size.x,w->pos.y+w->size.y);
         update=0;
      }

      UpdateMouse();
      bp=UpdateButtons();

      if (bp==b_ok)
         break;

      if (bp==b_cancel)
         break;

      if (bp==b_tex)
      {
         char texture[128];
         texture_t *tex;

         TexturePicker(NULL,texture);
         if (texture[0])
         {

            tex=ReadMIPTex(texture,1);
            if (tex)
            {
               GetTexColor(tex);

               col[0]=tex->colv[0]*col_max;
               col[1]=tex->colv[1]*col_max;
               col[2]=tex->colv[2]*col_max;
               update=1;
            }
         }
      }

      if (bp==b_scale)
      {
         max=col[0];
         for (i=1;i<3;i++)
            if (col[i]>max)
               max=col[i];
      
         if (max)
         {
            max=col_max/max;
            for (i=0;i<3;i++)
               col[i]*=max;
         }
         update=1;
      }

      for (i=0;i<3;i++)
      {
         if (bp==b_inc[i])
         {
            col[i]+=1;
            update=1;
         }
         if (bp==b_dec[i])
         {
            col[i]-=1;
            update=1;
         }
      }
   }

	/* Pop down the window (also copies back whats behind the window) */
	Q.num_popups--;
	QUI_PopDownWindow(POP_WINDOW_1+Q.num_popups,&TempBuf);

	/* Get rid of the button */
	RemoveButton(b_ok);
	RemoveButton(b_cancel);
   RemoveButton(b_tex);
   RemoveButton(b_scale);
   for (i=0;i<3;i++)
   {
      RemoveButton(b_inc[i]);
      RemoveButton(b_dec[i]);
   }
   PopButtons();

	/* refresh the correct portion of the screen */
	RefreshPart(w->pos.x,w->pos.y,w->pos.x+w->size.x,w->pos.y+w->size.y);        
	DrawMouse(mouse.x,mouse.y);

   if (bp==b_ok)
   {
      if (col_scale)
      {
         max=col[0];
         for (i=1;i<3;i++)
            if (col[i]>max)
               max=col[i];
   
         if (max)
         {
            max=col_max/max;
            for (i=0;i<3;i++)
               col[i]*=max;
         }
      }

      sprintf(buf,"%i %i %i",(int)col[0],(int)col[1],(int)col[2]);
      return 1;
   }
   else
      return 0;
}


static int PickCL(char *buf)
{
   EntityPicker(buf,CLASS_MODEL|CLASS_POINT);
   if (buf[0])
      return 1;
   else
      return 0;
}


static int GetChoice(char *buf,Q_key_t *k)
{
   int i;

   Popup_Init(mouse.x,mouse.y);
   for (i=0;i<k->x.ch.num_choices;i++)
      Popup_AddStr("%s\t%s",k->x.ch.choices[i].name,k->x.ch.choices[i].val);
   i=Popup_Display();
   Popup_Free();
   if (i!=-1)
   {
      strcpy(buf,k->x.ch.choices[i].val);
      return 1;
   }
   return 0;
}


static int GetFromList(char *buf,char *list)
{
   char temp[256];

   if (PickFromList(temp,list))
   {
      strcpy(buf,temp);
      return 1;
   }
   else
      return 0;
}



static void EditKey(entity_t *e,char *key,const char *v)
{
   class_t *c;
   int i;
   Q_key_t *k;
   char temp[256];

   strcpy(temp,v);
   if (!strcmp(key,"classname"))
   {
      if (PickCL(temp))
         SetKeyValue(e,key,temp);
      return;
   }

   c=FindClass(GetKeyValue(e,"classname"));
   if (!c)
   {
      HandleError("EditKey","Unknown key!");
      return;
   }

   for (i=0;i<c->num_keys;i++)
      if (!strcmp(c->keys[i].name,key))
         break;

   if (i==c->num_keys)
   {
      HandleError("EditKey","Unknown key!");
      return;
   }

   k=&c->keys[i];
   switch (k->type)
   {
   case KEY_HL_LIGHT:
      {
         float r,g,b,a;
         int i;

         i=sscanf(temp,"%f %f %f %f",&r,&g,&b,&a);
         if (i==1)
         {
            a=r;
            g=b=r=255;
         }
         else
         if (i==3)
         {
            if (g>r)
               a=g;
            else
               a=r;
            if (b>a)
               a=b;

            r*=255/a;
            g*=255/a;
            b*=255/a;
         }
         else
         if (i!=4)
         {
            r=g=b=255;
            a=300;
         }

         sprintf(temp,"%g %g %g",r,g,b);
         if (EditCol(temp,255,0))
         {
            sscanf(temp,"%f %f %f",&r,&g,&b);
            sprintf(temp,"%g %g %g %g",r,g,b,a);
            SetKeyValue(e,key,temp);
         }
      }
      break;
   case KEY_HL_COLOR:
      if (EditCol(temp,255,0))
         SetKeyValue(e,key,temp);
      break;

   case KEY_Q2_COLOR:
      if (EditCol(temp,100,1))
         SetKeyValue(e,key,temp);
      break;

   case KEY_CHOICE:
      if (GetChoice(temp,k))
         SetKeyValue(e,key,temp);
      break;

   case KEY_LIST:
      if (GetFromList(temp,k->x.list.name))
         SetKeyValue(e,key,temp);
      break;

   case KEY_TEXTURE:
      {
         char temp2[256];

         TexturePicker(temp,temp2);
         if (temp2[0])
            SetKeyValue(e,key,temp2);
      }
      break;
   }
}

static int CanEdit(entity_t *e,char *key)
{
   class_t *c;
   int i;

   if (!strcmp(key,"classname"))
      return 1;

   c=FindClass(GetKeyValue(e,"classname"));
   if (!c)
      return 0;

   for (i=0;i<c->num_keys;i++)
      if (!strcmp(c->keys[i].name,key))
         return 1;

   return 0;
}
