#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "edit.h"
#include "map.h"
#include "quest.h"
#include "s_error.h"
#include "s_gui.h"
#include "s_main.h"
#include "s_message.h"
#include "s_sys.h"


/* Basic commands */

static void C_exit(void)
{
   if (!Map_Close("Exit"))
      return;
   S_Exit();
}


static void C_map_new(void)
{
   if (!Map_Close("New map"))
      return;

   Map_New();
   Cmd_SetRedraw(2);
}

static void C_map_load(void)
{
   char name[MAX_PATH];

   if (!Map_Close("Load map"))
      return;

   if (!S_GUI_FileDialog("Load map","Load","*.map",name))
      return;

   S_NewMessage("Loading %s...",name);
   if (!Map_Load(name))
   {
      S_Error("map_load","Unable to load '%s'.",name);
   }

   Cmd_SetRedraw(2);
}

static void C_map_save_as(void)
{
   char name[MAX_PATH];

   if (!S_GUI_FileDialog("Save map","Save","*.map",name))
      return;

   if (Sys_Exists(name))
   {
		if (!S_GUI_YesNo("Save as","File exists: '%s'\n Overwrite?",name))
         return;
   }

   Map_SetName(name);
   if (!Map_Save())
      S_NewMessage("MAP saved as %s.",M.mapfilename);
}

static void C_map_save(void)
{
   Map_Save();
}


static cmd_reg_t cmds[]={
{"exit"       ,C_exit       ,{"File/Exit","escape"   },{CMD_NOUNDO},"Exit Quest."},
{"map.new"    ,C_map_new    ,{"File/New"             },{CMD_UNDO_CLEAR},"Start a new map."},
{"map.load"   ,C_map_load   ,{"File/Load..."         },{CMD_UNDO_CLEAR},"Load a map."},
{"map.save"   ,C_map_save   ,{"File/Save","control+s"},{CMD_NOUNDO},"Save the current map."},
{"map.save_as",C_map_save_as,{"File/Save as..."      },{CMD_NOUNDO},"Save the current map under a new name."},
};

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/basic","Fundamental commands (exit, new, load, etc.)."}};

PLUGIN_FOOT_CMDS

