#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "csg.h"
#include "edsel.h"
#include "geom.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"
#include "s_message.h"



static int WeldBrushes(bsel_t *br1,bsel_t *br2)
{
   int      i,j,k,l;
   face_t  *p1,*p2;
   int      w1,w2;
   brush_t *b1,*b2;

   b1=br1->Brush;
   b2=br2->Brush;

   if (b1->Entity!=b2->Entity) return 0;

   { // find the common face (could probably be made smarter)
      j=-1; // stop compiler warning
      for (i=0;i<b1->num_faces;i++)
      {
         p1=&b1->faces[i];
         for (j=0;j<b2->num_faces;j++)
         {
            p2=&b2->faces[j];
            if (p1->num_verts!=p2->num_verts)
               continue;

            if (fabs(p1->dist+p2->dist)>0.01)
               continue;

            if ((fabs(p1->normal.x+p2->normal.x)>0.01) ||
                (fabs(p1->normal.y+p2->normal.y)>0.01) ||
                (fabs(p1->normal.z+p2->normal.z)>0.01))
               continue;

            for (k=0;k<p1->num_verts;k++)
            {
               for (l=0;l<p1->num_verts;l++)
               {
                  if ((fabs(b1->verts[p1->verts[k]].x-b2->verts[p2->verts[l]].x)<0.01) &&
                      (fabs(b1->verts[p1->verts[k]].y-b2->verts[p2->verts[l]].y)<0.01) &&
                      (fabs(b1->verts[p1->verts[k]].z-b2->verts[p2->verts[l]].z)<0.01))
                     break;
               }
               if (l==p1->num_verts) break;
            }
            if (k==p1->num_verts)
               break;
         }
         if (j!=b2->num_faces)
         {
            break;
         }
      }
      if (i==b1->num_faces)
      {
//         NewMessage("No common face!");
         return 0;
      }
      w1=i;
      w2=j;
   }

   { // check resulting brush
   	vec3_t   delta;
   	vec3_t   center1, center2;

   	for (i=0;i<b1->num_faces;i++)
   	{
         if (i==w1)
            continue;

         p1=&b1->faces[i];
   		/* Create centerpoint for faces */
   		center1.x = center1.y = center1.z = 0;
   		for (j=0; j<p1->num_verts; j++)
   		{
   			center1.x += b1->verts[p1->verts[j]].x;
   			center1.y += b1->verts[p1->verts[j]].y;
   			center1.z += b1->verts[p1->verts[j]].z;
   		}
   		center1.x /= p1->num_verts;
   		center1.y /= p1->num_verts;
   		center1.z /= p1->num_verts;

   		for (j=0;j<b2->num_faces;j++)
   		{
            if (j==w2)
               continue;

            p2=&b2->faces[j];
   			/* Create centerpoint for other faces */
   			center2.x = center2.y = center2.z = 0;
   			for (k=0; k<p2->num_verts; k++)
   			{
   				center2.x += b2->verts[p2->verts[k]].x;
   				center2.y += b2->verts[p2->verts[k]].y;
   				center2.z += b2->verts[p2->verts[k]].z;
   			}
   			center2.x /= p2->num_verts;
   			center2.y /= p2->num_verts;
   			center2.z /= p2->num_verts;
   
   			/* Create vector from center1 to center2 */
   			delta.x = center2.x - center1.x;
   			delta.y = center2.y - center1.y;
   			delta.z = center2.z - center1.z;
   
   			/* If the dot prod of this delta and center2's normal
   				is less than 0 (with some epsilon), it's not convex */
   			if (DotProd(delta, p2->normal) < -.015)
            {
//               NewMessage("Result not convex!");
               return 0;
            }
   		}
   	}
   }

   { // check for identical faces with different textures
      for (i=0;i<b1->num_faces;i++)
      {
         p1=&b1->faces[i];
         for (j=0;j<b2->num_faces;j++)
         {
            p2=&b2->faces[j];

            if (p1->dist!=p2->dist)
               continue;

            if ((p1->normal.x!=p2->normal.x) ||
                (p1->normal.y!=p2->normal.y) ||
                (p1->normal.z!=p2->normal.z))
               continue;

            if (strcmp(p1->tex.name,p2->tex.name))
            {
/*               NewMessage(">1 texture/faces (%g %g %g) %g (%s and %s)",
                  p1->normal.x,p1->normal.y,p1->normal.z,
                  p1->dist,
                  p1->tex.name,p2->tex.name);*/
               return 0;
            }
         }
      }
   }

   { // weld the brushes!
      cbrush_t t;
      brush_t *b;

      t.num_faces=b1->num_faces+b2->num_faces-2;
      t.faces=Q_malloc(sizeof(cface_t)*t.num_faces);
      if (!t.faces)
      {
         S_Error("CSG Weld","Out of memory!");
         return -1;
      }

      j=0;
      for (i=0;i<b1->num_faces;i++)
      {
         if (i==w1)
            continue;
         CSG_FaceToCFace(&t.faces[j],b1,i);
         j++;
      }

      for (i=0;i<b2->num_faces;i++)
      {
         if (i==w2)
            continue;
         CSG_FaceToCFace(&t.faces[j],b2,i);
         j++;
      }

      b=CSG_BuildBrush(NULL,&t);
      if (!b)
      {
         S_Error("CSG Weld","Error building new brush!");
         return -1;
      }
      
      b->Entity = b1->Entity;
      b->Group = b1->Group;

      B_Link(b);

      br1->Brush=b;
      RemoveSelBrush(br2);

      B_Unlink(b2);
      B_Free(b2);
      B_Unlink(b1);
      B_Free(b1);
   }

   return 1;
}


static void C_weld(void)
{
   int i;
   int welds;
   bsel_t *br1,*br2;

   for (br1=M.sel.bsel;br1;br1=br1->Next)
   {
      if (br1->Brush->bt->type!=BRUSH_NORMAL)
      {
         S_Error("CSG Weld","Can only weld normal brushes!");
         return;
      }
   }

   i=0;
   welds=0;

   for (br1=M.sel.bsel;br1;br1=br1->Next)
   {
      br2=br1->Next;
restart_weld:
      for (;br2;br2=br2->Next)
      {
         if (br1==br2)
            continue;

         i=WeldBrushes(br1,br2);
         if (i==1)
         {
            welds++;
            br2=M.sel.bsel;
            goto restart_weld;
         }
         if (i==-1)
         {
            S_Error("CSG Weld","Error welding!");
            Cmd_SetRedraw(2);
            return;
         }
      }
   }

   S_NewMessage("%i welds done.",welds);
   if (welds)
      ClearSelVerts();

   Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"csg.weld",C_weld,{"Brush/CSG/Weld","w"},{},"CSG weld",NULL},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/csg/weld","CSG weld command."}};

PLUGIN_FOOT_CMDS

