#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "s_texture.h"
#include "q3_jpg.h"

#include "config.h"
#include "memory.h"
#include "s_message.h"


#include <jpeglib.h>


void quake3_load_jpg(const char *name,void *tex,const texture_t *t)
{
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   
   FILE *f;

   JSAMPROW row;

   unsigned char *data;

   int sx,sy;
   int asx,dsx,dsy;
   int scale;


   f=fopen(name,"rb");
   if (!f)
   {
      S_NewMessage("quake3.load_jpg: Can't open '%s'!",name);
      return;
   }

   cinfo.err=jpeg_std_error(&jerr);
   jpeg_create_decompress(&cinfo);

   jpeg_stdio_src(&cinfo,f);
   jpeg_read_header(&cinfo,TRUE);

   dsx=sx=cinfo.image_width;
   dsy=sy=cinfo.image_height;

   scale=0;
   if (dsx>cv_quake3_scale_above->i && dsy>cv_quake3_scale_above->i) dsx/=2,dsy/=2,scale++;
   while (scale<3 && (dsx>cv_quake3_scale_below->i || dsy>cv_quake3_scale_below->i))
      dsx/=2,dsy/=2,scale++;

#ifdef COUNT_TEXEL
   count_texel(dsx,dsy);
   jpeg_destroy_decompress(&cinfo);
   fclose(f);
   return;
#endif

   cinfo.scale_num=1;
   cinfo.scale_denom=1<<scale;

   cinfo.quantize_colors=FALSE;
   cinfo.out_color_space=JCS_RGB;

/* Make things faster */
   cinfo.dct_method=JDCT_FLOAT;
   cinfo.do_fancy_upsampling=FALSE;

   jpeg_start_decompress(&cinfo);

   dsx=cinfo.output_width;
   dsy=cinfo.output_height;

   for (asx=dsx*3;asx&3;asx++); /* TODO: this a an OpenGL specific hack */

   data=Q_malloc(asx*dsy);
   if (!data)
   {
      jpeg_destroy_decompress(&cinfo);
      fclose(f);
      S_NewMessage("quake3.load_jpg: Out of memory!");
      return;
   }

   while (cinfo.output_scanline<cinfo.output_height)
   {
      row=&data[cinfo.output_scanline*asx];
      jpeg_read_scanlines(&cinfo,&row,1);
   }

   jpeg_finish_decompress(&cinfo);
   jpeg_destroy_decompress(&cinfo);

   S_T_AddData24(tex,sx,sy,dsx,dsy,data);

   Q_free(data);

   fclose(f);
}


void quake3_load_jpg_size(const char *name,void *tex,const texture_t *t)
{
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   
   FILE *f;

   f=fopen(name,"rb");
   if (!f)
      return;

   cinfo.err=jpeg_std_error(&jerr);
   jpeg_create_decompress(&cinfo);

   jpeg_stdio_src(&cinfo,f);
   jpeg_read_header(&cinfo,TRUE);

   S_T_AddSize(tex,cinfo.image_width,cinfo.image_height);

   jpeg_destroy_decompress(&cinfo);
   fclose(f);
}

