#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gtk/gtk.h>

#include "g_gui.h"

#include "s_error.h"

#include "g_menu.h"


/* TODO: think of some better way to do this */
#define ARG_UNUSED(x) x __attribute__ ((unused))


static gint dialog_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),ARG_UNUSED(gpointer data))
{
   return TRUE;
}

static int dialog_selected;

static void dialog_click(ARG_UNUSED(GtkWidget *b),int i)
{
   dialog_selected=i;
   gtk_main_quit();
}

int S_GUI_Dialog(const char *title,const char *text, ...)
{
   va_list args;
   const char *a;
   int i;
   
   GtkWindow *win;
   GtkBox *vb;
   GtkWidget *b;
   GtkContainer *bb;
   

   win=GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
   gtk_window_set_title(win,title);
   gtk_window_set_position(win,GTK_WIN_POS_MOUSE);
   gtk_window_set_modal(win,TRUE);
   gtk_window_set_transient_for(win,S_menu_window);
   gtk_container_set_border_width(GTK_CONTAINER(win),10);
   gtk_signal_connect(GTK_OBJECT(win),"delete_event",GTK_SIGNAL_FUNC(dialog_delete),NULL);

   vb=GTK_BOX(gtk_vbox_new(FALSE,4));
   gtk_container_add(GTK_CONTAINER(win),GTK_WIDGET(vb));

   b=gtk_label_new(text);
   gtk_box_pack_start(vb,b,0,0,0);
   gtk_widget_show(b);

   bb=GTK_CONTAINER(gtk_hbutton_box_new());
   gtk_button_box_set_layout(GTK_BUTTON_BOX(bb),GTK_BUTTONBOX_EDGE);
   gtk_button_box_set_spacing(GTK_BUTTON_BOX(bb),4);
   gtk_box_pack_start(vb,GTK_WIDGET(bb),0,0,0);

   va_start(args,text);

   for (i=0,a=va_arg(args,char *);a;a=va_arg(args,char *),i++)
   {
      b=gtk_button_new_with_label(a);
      gtk_signal_connect(GTK_OBJECT(b),"clicked",GTK_SIGNAL_FUNC(dialog_click),i);
      gtk_container_add(bb,b);
      gtk_widget_show(b);
      if (!i) gtk_widget_grab_focus(b);
   }

   va_end(args);

   gtk_widget_show(GTK_WIDGET(bb));
   gtk_widget_show(GTK_WIDGET(vb));
   gtk_widget_show(GTK_WIDGET(win));

   gtk_main();
   gtk_widget_destroy(GTK_WIDGET(win));

   return dialog_selected;
}


void S_GUI_DialogOK(const char *title,const char *format, ...)
{
   char buf[1024];
   va_list args;

   va_start(args,format);
   vsnprintf(buf,sizeof(buf)-1,format,args);
   va_end(args);
   
   S_GUI_Dialog(title,buf,"OK",NULL);
}

int S_GUI_YesNo(const char *title,const char *format, ...)
{
   char buf[1024];
   va_list args;

   va_start(args,format);
   vsnprintf(buf,sizeof(buf)-1,format,args);
   va_end(args);
   
   return !(S_GUI_Dialog(title,buf,"Yes","No",NULL));
}


static void file_ok(ARG_UNUSED(GtkWidget *w),int *ok)
{
   *ok=1;
   gtk_main_quit();
}

static void file_cancel(ARG_UNUSED(GtkWidget *w),int *ok)
{
   *ok=0;
   gtk_main_quit();
}

static gint file_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),int *ok)
{
   *ok=0;
   gtk_main_quit();
   return TRUE;
}

int S_GUI_FileDialog(const char *title,const char *button,
                     const char *wildcard,char *name)
{
   GtkFileSelection *fs;
   int ok=0;

   fs=GTK_FILE_SELECTION(gtk_file_selection_new(title));
   gtk_window_set_modal(GTK_WINDOW(fs),TRUE);
   gtk_window_set_position(GTK_WINDOW(fs),GTK_WIN_POS_MOUSE);

   gtk_signal_connect(GTK_OBJECT(fs->ok_button),"clicked",GTK_SIGNAL_FUNC(file_ok),&ok);
   gtk_signal_connect(GTK_OBJECT(fs->cancel_button),"clicked",GTK_SIGNAL_FUNC(file_cancel),&ok);
   gtk_signal_connect(GTK_OBJECT(fs),"delete_event",GTK_SIGNAL_FUNC(file_delete),&ok);
   gtk_label_set_text(GTK_LABEL(GTK_BIN(fs->ok_button)->child),button);
   gtk_widget_show(GTK_WIDGET(fs));

   gtk_main();
   if (ok)
      strcpy(name,gtk_file_selection_get_filename(fs));
   gtk_widget_destroy(GTK_WIDGET(fs));

   return ok;
}


static gint pop_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),int *ok)
{
   *ok=0;
   gtk_main_quit();
   return TRUE;
}

static void pop_ok_clicked(ARG_UNUSED(GtkWidget *w),int *ok)
{
   *ok=1;
   gtk_main_quit();
}

static void pop_cancel_clicked(ARG_UNUSED(GtkWidget *w),int *ok)
{
   *ok=0;
   gtk_main_quit();
}

static void pop_next(ARG_UNUSED(GtkWidget *w),GtkContainer *c)
{
   gtk_container_focus(c,GTK_DIR_TAB_FORWARD);
}

int S_GUI_PopEntity(const char *title,char **keys,char **values,int num_keys)
{
   GtkWindow *win;
   GtkEntry  **ent;
   GtkWidget *b;
   GtkBox   *vb,*hb;
   
   int ok=0;
   int i;
   

   ent=malloc(sizeof(GtkEntry *)*(num_keys));
   if (!ent)
   {
      S_Error("S_GUI_PopEntity","Out of memory!");
      return 0;
   }

   win=GTK_WINDOW(gtk_window_new(GTK_WINDOW_DIALOG));
   gtk_window_set_modal(win,TRUE);
   gtk_window_set_position(win,GTK_WIN_POS_MOUSE);
   gtk_window_set_title(win,title);
   gtk_signal_connect(GTK_OBJECT(win),"delete_event",GTK_SIGNAL_FUNC(pop_delete),&ok);
   gtk_container_set_border_width(GTK_CONTAINER(win),8);

   vb=GTK_BOX(gtk_vbox_new(FALSE,4));
   gtk_container_add(GTK_CONTAINER(win),GTK_WIDGET(vb));

   for (i=0;i<num_keys;i++)
   {
      ent[i]=GTK_ENTRY(gtk_entry_new());
      gtk_entry_set_text(ent[i],values[i]);

      gtk_signal_connect(GTK_OBJECT(ent[i]),"activate",GTK_SIGNAL_FUNC(pop_next),vb);
      
      b=gtk_label_new(keys[i]);
      gtk_misc_set_alignment(GTK_MISC(b),0.0,0.5);

      gtk_box_pack_start(vb,b,0,0,0);
      gtk_box_pack_start(vb,GTK_WIDGET(ent[i]),0,0,0);

      gtk_widget_show(b);
      gtk_widget_show(GTK_WIDGET(ent[i]));

      if (!i) gtk_widget_grab_focus(GTK_WIDGET(ent[i]));
   }

   b=gtk_hseparator_new();
   gtk_widget_show(b);
   gtk_box_pack_start(vb,b,0,0,0);

   hb=GTK_BOX(gtk_hbox_new(FALSE,0));
   gtk_box_pack_start(vb,GTK_WIDGET(hb),0,0,0);
   gtk_widget_show(GTK_WIDGET(hb));

   b=gtk_button_new_with_label("OK");
   gtk_signal_connect(GTK_OBJECT(b),"clicked",GTK_SIGNAL_FUNC(pop_ok_clicked),&ok);
   gtk_box_pack_start(hb,b,0,0,0);
   gtk_widget_show(b);
   
   b=gtk_button_new_with_label("Cancel");
   gtk_signal_connect(GTK_OBJECT(b),"clicked",GTK_SIGNAL_FUNC(pop_cancel_clicked),&ok);
   gtk_box_pack_start(hb,b,0,0,0);
   gtk_widget_show(b);

   gtk_widget_show(GTK_WIDGET(vb));
   gtk_widget_show(GTK_WIDGET(win));

   gtk_main();

   if (ok)
      for (i=0;i<num_keys;i++)
         strcpy(values[i],gtk_entry_get_text(ent[i]));

   gtk_widget_destroy(GTK_WIDGET(win));

   free(ent);

   return ok;
}


#include "config.h"

/* TODO: figure out a way to do this without using 'internal' functions */
void S_LoadWinPosition(GtkWindow *win,const char *name)
{
   char buf[128];
   const char *c;
   int x,y,sx,sy;

   sprintf(buf,"gtk.win.%s",name);
   c=Cfg_Str(buf);
   if (c)
   {
      if (sscanf(c,"%i %i %i %i",&x,&y,&sx,&sy)==4)
      {
         gtk_window_set_default_size(win,sx,sy);
         gtk_window_reposition(win,x,y);
      }
   }
}

void S_SaveWinPosition(GtkWindow *win,const char *name)
{
   char buf[128];
   const char *c;
   config_var_t *cv;
   int x,y,sx,sy;
   
   sprintf(buf,"gtk.win.%s",name);
   c=Cfg_Str(buf);
   if (c && !strcmp(c,"-"))
      return;

   cv=Cfg_DefVar(buf,"",NULL);
   gdk_window_get_root_origin(GTK_WIDGET(win)->window,&x,&y);
   gdk_window_get_size(GTK_WIDGET(win)->window,&sx,&sy);
   sprintf(buf,"%i %i %i %i",x,y,sx,sy);
   Cfg_SetStr(cv,buf);
}

