#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gtk/gtk.h>

#undef TRUE
#undef FALSE

#include "defines.h"
#include "types.h"

#include "s_gui.h"

#include "config.h"
#include "memory.h"
#include "s_error.h"


#define TI_DATA_FIELD "quest-ti"


typedef struct cv_tree_s
{
   char *name,*fname;
   struct cv_tree_s *next,*child,*parent;
} cv_tree_t;

static cv_tree_t *ctree;


static cv_tree_t *TreeFind(const char *str,cv_tree_t *r)
{
   cv_tree_t *rp;
   char fname[128];

restart:
   for (rp=r;r;rp=r,r=r->next)
      if (!strncmp(r->name,str,strlen(r->name)))
         break;
   if (!r)
   {
      r=rp->next=Q_malloc(sizeof(cv_tree_t));
      if (!r) return NULL;
      memset(r,0,sizeof(cv_tree_t));
      r->parent=rp->parent;
      r->name=Q_strdup(str);
      if (!r->name) return NULL;
      if (strchr(r->name,'.'))
         *strchr(r->name,'.')=0;

      strcpy(fname,r->parent->fname);
      if (fname[0])
         strcat(fname,".");
      strcat(fname,r->name);
      r->fname=Q_strdup(fname);
      if (!r->fname) return NULL;
   }
   str=strchr(str,'.');
   if (!str)
      return r;
   str++;
   if (!r->child)
   {
      r->child=Q_malloc(sizeof(cv_tree_t));
      if (!r->child) return NULL;
      memset(r->child,0,sizeof(cv_tree_t));
      r->child->parent=r;
      r->child->name=Q_strdup(str);
      if (!r->child->name) return NULL;
      if (strchr(r->child->name,'.'))
         *strchr(r->child->name,'.')=0;
         
      strcpy(fname,r->fname);
      if (fname[0])
         strcat(fname,".");
      strcat(fname,r->child->name);
      r->child->fname=Q_strdup(fname);
      if (!r->child->fname) return NULL;
   }
   r=r->child;
   goto restart;
/*   return TreeFind(str,r->child);*/
}

static cv_tree_t *TreeAdd(const char *c)
{
   char buf[128];

   buf[0]='.';
   strcpy(&buf[1],c);
   *strrchr(buf,'.')=0;
   return TreeFind(buf,ctree);
}

static void TreeFree(cv_tree_t *t)
{
   cv_tree_t *tn;
   for (;t;t=tn)
   {
      if (t->child) TreeFree(t->child);
      if (t->name) Q_free(t->name);
      if (t->fname) Q_free(t->fname);
      tn=t->next;
      Q_free(t);
   }
}


static void MakeTree(void)
{
   int i;

   ctree=Q_malloc(sizeof(cv_tree_t));
   if (!ctree)
   {
      S_Error("Preferences","Out of memory!");
      return;
   }
   memset(ctree,0,sizeof(cv_tree_t));
   ctree->name=Q_strdup("");
   ctree->fname=Q_strdup("");
   if (!ctree->name || !ctree->fname)
   {
      S_Error("Preferences","Out of memory!");
      return;
   }

   for (i=0;i<num_cvars;i++)
   {
      if (!TreeAdd(cvars[i]->name))
      {
         S_Error("Preferences","Out of memory!");
         return;
      }
   }
}


static GtkTree *TreeGUI(cv_tree_t *r)
{
   GtkWidget *ti;
   GtkTree *t=GTK_TREE(gtk_tree_new());
   for (;r;r=r->next)
   {
      if (r->name[0])
         ti=gtk_tree_item_new_with_label(r->name);
      else
         ti=gtk_tree_item_new_with_label("Quest");
      gtk_object_set_data(GTK_OBJECT(ti),TI_DATA_FIELD,r->fname);
      gtk_tree_append(t,ti);
      gtk_widget_show(ti);
      if (r->child)
         gtk_tree_item_set_subtree(GTK_TREE_ITEM(ti),GTK_WIDGET(TreeGUI(r->child)));
   }
   return t;
}


static GtkScrolledWindow *pane_sw;
static GtkBox *pane_box;


static void s_changed(GtkEntry *e,config_var_t *c)
{
   Cfg_SetStr(c,gtk_entry_get_text(e));
}


static void make_pane(const char *base)
{ /* TODO: try to make this look good */
   GtkWidget *l;

   if (pane_box)
   {
      gtk_widget_destroy(GTK_WIDGET(pane_box));
      pane_box=NULL;
   }

   pane_box=GTK_BOX(gtk_vbox_new(FALSE,2));
   gtk_container_set_border_width(GTK_CONTAINER(pane_box),4);

   if (base[0])
      l=gtk_label_new(base);
   else
      l=gtk_label_new("Quest");
   gtk_box_pack_start(pane_box,l,FALSE,FALSE,2);
   gtk_widget_show(l);

   {
      int i;
      int num;
      config_var_t *c,**cv;
      GtkEntry *e;
      GtkTable *t;

      num=0;
      for (i=num_cvars,cv=cvars;i;i--,cv++)
      {
         c=*cv;
         if (strncmp(base,c->name,strlen(base)) ||
             strchr(c->name+strlen(base)+1,'.'))
            continue;
         num++;
      }

      t=GTK_TABLE(gtk_table_new(3*num,2,FALSE));
      gtk_box_pack_start(pane_box,GTK_WIDGET(t),FALSE,FALSE,2);
      gtk_widget_show(GTK_WIDGET(t));
      
      num=0;
      for (i=num_cvars,cv=cvars;i;i--,cv++)
      {
         c=*cv;
         if (strncmp(base,c->name,strlen(base)) ||
             strchr(c->name+strlen(base)+1,'.'))
            continue;

         l=gtk_hseparator_new();
         gtk_table_attach(t,l,0,2,num*3,num*3+1,
            GTK_FILL|GTK_EXPAND|GTK_SHRINK,0,
            2,2);
         gtk_widget_show(l);

         l=gtk_label_new(c->name);
         gtk_misc_set_alignment(GTK_MISC(l),0.0,0.5);
         gtk_table_attach(t,l,0,1,num*3+1,num*3+2,
            GTK_FILL|GTK_SHRINK,0,
            2,2);
         gtk_widget_show(l);

         e=GTK_ENTRY(gtk_entry_new());
         gtk_entry_set_text(e,c->s);
         gtk_signal_connect(GTK_OBJECT(e),"changed",GTK_SIGNAL_FUNC(s_changed),c);
         gtk_table_attach(t,GTK_WIDGET(e),0,1,num*3+2,num*3+3,
            GTK_FILL|GTK_SHRINK,0,
            2,2);
         gtk_widget_show(GTK_WIDGET(e));

         if (c->desc)
         {
            l=gtk_text_new(NULL,NULL);
            gtk_text_insert(GTK_TEXT(l),NULL,NULL,NULL,c->desc,-1);
            gtk_text_set_word_wrap(GTK_TEXT(l),TRUE);
            gtk_widget_set_usize(l,1,32);
            gtk_table_attach(t,l,1,2,num*3+2,num*3+3,
               GTK_FILL|GTK_EXPAND|GTK_SHRINK,GTK_FILL|GTK_EXPAND|GTK_SHRINK,
               2,2);
            gtk_widget_show(l);
         }

         num++;
      }
   }
   
   gtk_scrolled_window_add_with_viewport(pane_sw,GTK_WIDGET(pane_box));
   gtk_widget_show(GTK_WIDGET(pane_box));
}


static void s_ok_clicked(ARG_UNUSED(GtkButton *b),ARG_UNUSED(gpointer data))
{
   gtk_main_quit();
}

static gint s_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),ARG_UNUSED(gpointer data))
{
   gtk_main_quit();
   return TRUE;
}

static void s_sel(GtkTree *t,ARG_UNUSED(gpointer data))
{
   GList *sel;
   char *c;

   sel=GTK_TREE_SELECTION(t);
   c=gtk_object_get_data(GTK_OBJECT(sel->data),TI_DATA_FIELD);
   if (c)
      make_pane(c);
   else
      make_pane("");
}


static void S_Preferences(void)
{
   GtkWindow *win;
   GtkBox *vb;

   MakeTree();

   win=GTK_WINDOW(gtk_window_new(GTK_WINDOW_DIALOG));
   gtk_window_set_modal(win,TRUE);
   gtk_window_set_position(win,GTK_WIN_POS_MOUSE);
   gtk_window_set_title(win,"Preferences");
   gtk_signal_connect(GTK_OBJECT(win),"delete_event",GTK_SIGNAL_FUNC(s_delete),NULL);
   gtk_container_set_border_width(GTK_CONTAINER(win),4);
   gtk_window_set_default_size(win,400,300);

   vb=GTK_BOX(gtk_vbox_new(FALSE,2));

   gtk_container_add(GTK_CONTAINER(win),GTK_WIDGET(vb));

   {
      GtkPaned *p;
      GtkTree *t;
      GtkWidget *s;

      p=GTK_PANED(gtk_hpaned_new());
      gtk_box_pack_start(vb,GTK_WIDGET(p),TRUE,TRUE,4);
      gtk_paned_set_position(p,128);
      gtk_widget_show(GTK_WIDGET(p));

      s=gtk_scrolled_window_new(NULL,NULL);
      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(s),
         GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
      gtk_container_set_border_width(GTK_CONTAINER(s),4);

      t=TreeGUI(ctree);
      gtk_tree_set_selection_mode(t,GTK_SELECTION_BROWSE);
      gtk_widget_show(GTK_WIDGET(t));
      gtk_signal_connect(GTK_OBJECT(t),"selection_changed",GTK_SIGNAL_FUNC(s_sel),NULL);

      gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(s),
         GTK_WIDGET(t));

      gtk_widget_show(s);
      gtk_paned_add1(p,s);

      pane_sw=GTK_SCROLLED_WINDOW(gtk_scrolled_window_new(NULL,NULL));
      gtk_scrolled_window_set_policy(pane_sw,
         GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
      gtk_container_set_border_width(GTK_CONTAINER(pane_sw),4);
      gtk_widget_show(GTK_WIDGET(pane_sw));
      gtk_paned_add2(p,GTK_WIDGET(pane_sw));
      pane_box=NULL;
   }

   {
      GtkBox *bb;
      GtkWidget *b;
      
      bb=GTK_BOX(gtk_hbox_new(FALSE,2));
      gtk_box_pack_start(vb,GTK_WIDGET(bb),FALSE,FALSE,4);

      b=gtk_button_new_with_label("Done");
      gtk_signal_connect(GTK_OBJECT(b),"clicked",GTK_SIGNAL_FUNC(s_ok_clicked),NULL);
      gtk_box_pack_start(bb,b,FALSE,FALSE,4);
      gtk_widget_show(b);

      gtk_widget_show(GTK_WIDGET(bb));
   }
      
   gtk_widget_show(GTK_WIDGET(vb));
   gtk_widget_show(GTK_WIDGET(win));


   gtk_main();
   
   gtk_widget_destroy(GTK_WIDGET(win));

   TreeFree(ctree);
}


#include "cmds.h"

static void C_prefs(void)
{
   S_Preferences();
   Cmd_SetRedraw(2);
}

static cmd_reg_t cmds[]=
{
{"preferences",C_prefs,{"File/Preferences...","control+p"},{},"Brings up the preferences dialog, letting you configure Quest."},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/gtk/preferences","The preferences dialog."}};

PLUGIN_FOOT_CMDS

