#ifndef S_INPUT_H
#define S_INPUT_H

/*
Each input-event thing has 8 bits (should be enough) that tell what
event it was (which key, which mouse button, etc.) and it might also
have a combination of the KEYM flags, which tells you what modifiers
are being held down.

WARNING: Due to different keymaps and other evil things, not all keys
can be combined with modifiers. For now, a-z, 0-9 (keypad), f1-f12,
escape, tab, enter, space, and the arrows and their related keys are safe.
*/

#define KEYM_SHIFT   0x100
#define KEYM_ALT     0x200
#define KEYM_CONTROL 0x400


#define  KEY_NONE     0x00
#define  KEY_MASK     0xff

/* Mouse stuff */
#define  KEY_BUTTON1  0xf0   /* left mouse button */
#define  KEY_BUTTON2  0xf1   /* right mouse button */
#define  KEY_BUTTON3  0xf2   /* middle mouse button */
/* TODO: mouse wheel support? */


/* All the keys */
#define  KEY_ESCAPE   0x01
#define  KEY_ENTER    0x02
#define  KEY_SPACE    0x03
#define  KEY_TAB      0x04
#define  KEY_BACKSPC  0x05

/* Not that useful, but might as well
include them. */
#if TODO /* TODO: fix these */
#define  KEY_COLON    0x06
#define  KEY_QUOTE    0x07
#define  KEY_TILDE    0x08
#define  KEY_PIPE     0x09
#define  KEY_COMMA    0x0a
#define  KEY_PERIOD   0x0b
#define  KEY_Q_MARK   0x0c
#define  KEY_CENTER   0x0d

#define  KEY_LF_BRACK 0x0e
#define  KEY_RT_BRACK 0x0f
#endif

#define  KEY_UP       0x10
#define  KEY_DOWN     0x11
#define  KEY_LEFT     0x12
#define  KEY_RIGHT    0x13

#define  KEY_INSERT   0x14
#define  KEY_DELETE   0x15

#define  KEY_HOME     0x16
#define  KEY_END      0x17

#define  KEY_PAGEUP   0x18
#define  KEY_PAGEDOWN 0x19

#define  KEY_MINUS    0x1a
#define  KEY_PLUS     0x1b

/* Letters */
#define  KEY_A        0x20
#define  KEY_B        0x21
#define  KEY_C        0x22
#define  KEY_D        0x23
#define  KEY_E        0x24
#define  KEY_F        0x25
#define  KEY_G        0x26
#define  KEY_H        0x27
#define  KEY_I        0x28
#define  KEY_J        0x29
#define  KEY_K        0x2a
#define  KEY_L        0x2b
#define  KEY_M        0x2c
#define  KEY_N        0x2d
#define  KEY_O        0x2e
#define  KEY_P        0x2f
#define  KEY_Q        0x30
#define  KEY_R        0x31
#define  KEY_S        0x32
#define  KEY_T        0x33
#define  KEY_U        0x34
#define  KEY_V        0x35
#define  KEY_W        0x36
#define  KEY_X        0x37
#define  KEY_Y        0x38
#define  KEY_Z        0x39

/* Digits */
#define  KEY_0        0x40
#define  KEY_1        0x41
#define  KEY_2        0x42
#define  KEY_3        0x43
#define  KEY_4        0x44
#define  KEY_5        0x45
#define  KEY_6        0x46
#define  KEY_7        0x47
#define  KEY_8        0x48
#define  KEY_9        0x49

/* Function Keys */
#define  KEY_F1       0x50
#define  KEY_F2       0x51
#define  KEY_F3       0x52
#define  KEY_F4       0x53
#define  KEY_F5       0x54
#define  KEY_F6       0x55
#define  KEY_F7       0x56
#define  KEY_F8       0x57
#define  KEY_F9       0x58
#define  KEY_F10      0x59
#define  KEY_F11      0x5a
#define  KEY_F12      0x5b


/* These are defined in s_input.c */
int Key_Parse(const char *str); /* Parses a string (like '{modifier+}*key'). */
const char *Key_Name(int k);

/* These are system dependent */
void S_Key_Bind(const char *key,const char *cmd);

#endif

