#include <stdio.h>
#include <stdlib.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "camera.h"
#include "csg.h"
#include "edsel.h"
#include "geom.h"
#include "map.h"
#include "quest.h"
#include "s_3d.h"
#include "s_error.h"
#include "s_input.h"
#include "s_int.h"
#include "s_message.h"
#include "s_trans.h"

/* TODO
valid only in 2d-viewports?
draw line
*/

static void C_clip(void)
{
	s_event_t e;
	vec3_t v1,v2;
	vec3_t dv,du;
	vec3_t m;
	plane_t p;

	Move(S_cur_vport,MOVE_FORWARD,&du,1);

	if (!M.sel.num_bselected)
	{
		S_NewMessage("Must select at least one brush!");
		return;
	}
	
	S_I_Setup(&e,S_I_CLICK,"Select the first point");
	while (1)
	{
		S_I_Wait(&e);
		if (e.type==S_I_CLICK)
		{
			if (e.key==KEY_BUTTON1 && S_R_GetMousePos(&m) && S_R_Trans_SVert(m,&v1))
				break;
			if (e.key==KEY_BUTTON2)
			{
				S_I_Done();
				S_NewMessage("Operation canceled.");
				return;
			}
		}
	}
	S_I_Done();
/* TODO: draw line */
	S_I_Setup(&e,S_I_CLICK,"Select the second point");
	while (1)
	{
		S_I_Wait(&e);
		if (e.type==S_I_CLICK)
		{
			if (e.key==KEY_BUTTON1 && S_R_GetMousePos(&m) && S_R_Trans_SVert(m,&v2))
				break;
			if (e.key==KEY_BUTTON2)
			{
				S_I_Done();
				S_NewMessage("Operation canceled.");
				return;
			}
		}
	}
	S_I_Done();

	SnapVecToGrid(&v1);
	SnapVecToGrid(&v2);

	VEC_Sub(dv,v1,v2);
	CrossProd(du,dv,&p.n);
	Normalize(&p.n);
	p.dist=DotProd(p.n,v1);
	{
		bsel_t *bs;
		brush_t *b,*nb;
		cbrush_t cf,cb;
		cface_t f;
		int i=0;

		for (bs=M.sel.bsel;bs;bs=bs->Next)
		{
			b=bs->Brush;
			if (!CSG_BrushToCBrush(&cf,b))
				continue; /* ignore it */
			if (!CSG_FaceToCFace(&f,b,0))
				continue; /* ignore it */
			f.p=p;
			switch (CSG_SplitBrushWithFace(&cf,f,&cb))
			{
			case 0:
				nb=CSG_BuildBrush(NULL,&cf);
				nb->Group=b->Group;
				nb->Entity=b->Entity;
				B_Link(nb);
				nb=CSG_BuildBrush(NULL,&cb);
				nb->Group=b->Group;
				nb->Entity=b->Entity;
				B_Link(nb);
				B_Unlink(b);
				B_Free(b);
				break;
			case -1:
			case -2:
				break;
			default:
				S_Error("brush.clip_plane","Error clipping brush!");
				break;
			}
			i++;
		}
		S_NewMessage(i!=1?"%i brushes split.":"%i brush split.",i);
	}
	ClearSelBrushes();
	ClearSelFaces();
	Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"brush.clip_plane",C_clip,{"Brush/Clip with plane","alt+c"},{},"Clip the selected brushes with a plane"},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/clip_plane","Clipping plane operations"}};

PLUGIN_FOOT_CMDS

