#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "brush.h"
#include "csg.h"
#include "edsel.h"
#include "error.h"
#include "memory.h"
#include "quest.h"
#include "fe/message.h"


static void C_join(void)
{
	int i;
	cbrush_t A,B,C;
	map_uid_t EntityRef;
	brush_t *b,*b1,*b2;

	if (M.sel.num_bselected!=2)
	{
		S_NewMessage("CSG Join: Must have exactly 2 brushes selected!");
		return;
	}

	b1=M.sel.bsel->Brush;
	b2=M.sel.bsel->Next->Brush;

	if (b1->bt->type!=BRUSH_NORMAL || b2->bt->type!=BRUSH_NORMAL)
	{
		S_NewMessage("CSG Join: Can only use normal brushes in CSG operations!");
		return;
	}

	EntityRef = b1->Entity;

	ClearSelVerts();

	if (!CSG_BrushToCBrush(&A,b1) || !CSG_BrushToCBrush(&B,b2))
	{
		Error_Add("Out of memory!");
		return;
	}

	C.num_faces = 0;
	C.faces = Q_malloc(sizeof(cface_t)*(A.num_faces+B.num_faces));
	if (!C.faces)
	{
		Error_Add("Out of memory!");
		return;
	}


/*
 For each face of b1 and b2, we must ask whether or not to make it
 a face of C.  One way that might work could be to treat each face
 of a brush as a plane and attempt to split the faces of the other
 brush with it.  Only if all the other faces are coplanar or entirely
 behind the plane should we add it to C.  Coplanars of one should be
 kept, and coplanars of the other should be tossed.


This doesn't always work, e.g.

+-----+A
|     |
| b2  |
|     |
+-----+B
|     \
| b1   \
|       \
+--------+C

The plane AB splits b1, and BC splits b2, so we end up with no plane on that
side. The correct way to solve this would probably be to try to form all
possible planes from triplets of vertices (two from one brush, one from the
other), and add the ones that aren't already in the new brush and that
wouldn't split b1 or b2.

TODO: is there a way to optimize the search for valid vertex triplets?

*/

	for (i=0; i<A.num_faces; i++)
		if (CSG_BrushSide(&B,A.faces[i].p)==-1)
			CSG_FaceToCFace(&C.faces[C.num_faces++],b1,i);

	for (i=0; i<B.num_faces; i++)
		if (CSG_BrushSide(&A,B.faces[i].p)==-1)
			CSG_FaceToCFace(&C.faces[C.num_faces++],b2,i);

	{
		int i,j,k;
		cface_t f;

		if (!CSG_FaceToCFace(&f,b1,0))
		{
			Error_Add("Out of memory!");
			return;
		}
		for (i=0;i<b1->num_verts;i++)
		{
			for (j=i+1;j<b1->num_verts;j++)
			{
				for (k=0;k<b2->num_verts;k++)
				{
					f.planepts[0]=b1->verts[i];
					f.planepts[1]=b1->verts[j];
					f.planepts[2]=b2->verts[k];

					if (CSG_CalcPlane(&f)==-1)
						continue;

					if (CSG_BrushSide(&A,f.p)!=-1)
						continue;
					if (CSG_BrushSide(&B,f.p)!=-1)
						continue;

					C.faces=Q_realloc(C.faces,sizeof(cface_t)*(C.num_faces+1));
					if (!C.faces)
					{
						Error_Add("Out of memory!");
						return;
					}
					C.faces[C.num_faces++]=f;
					if (!CSG_FaceToCFace(&f,b1,0))
					{
						Error_Add("Out of memory!");
						return;
					}
				}
			}
		}
		for (i=0;i<b1->num_verts;i++)
		{
			for (j=0;j<b2->num_verts;j++)
			{
				for (k=j+1;k<b2->num_verts;k++)
				{
					f.planepts[0]=b1->verts[i];
					f.planepts[1]=b2->verts[j];
					f.planepts[2]=b2->verts[k];

					if (CSG_CalcPlane(&f)==-1)
						continue;

					if (CSG_BrushSide(&A,f.p)!=-1)
						continue;
					if (CSG_BrushSide(&B,f.p)!=-1)
						continue;

					C.faces=Q_realloc(C.faces,sizeof(cface_t)*(C.num_faces+1));
					if (!C.faces)
					{
						Error_Add("Out of memory!");
						return;
					}
					C.faces[C.num_faces++]=f;
					if (!CSG_FaceToCFace(&f,b1,0))
					{
						Error_Add("Out of memory!");
						return;
					}
				}
			}
		}
		CSG_FreeFace(&f);
	}

	CSG_FreeBrush(&A);
	CSG_FreeBrush(&B);

	b=CSG_BuildBrush(NULL,&C);

	if (!b)
	{
		Error_Add("Error building brush!");
		return;
	}

	ClearSelBrushes();

	b->Entity=EntityRef;
	b->Group=b1->Group;
	B_Link(b);

	AddSelBrush(b);
}


static func_t F_csg_join=
{"csg.join",FUNC_VOID,{fv:C_join},NULL,NULL,0,
{"CSG join.",
"This function takes two brushes and attempts to create a single "
"convex brush that is the union of the two.\n"
"-----\n"
"originally by Gyro Gearloose"}
};

static func_t *funcs[]={&F_csg_join};


#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"csg.join","CSG join command."}};

PLUGIN_FOOT_FUNCS

