#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "brush.h"
#include "csg.h"
#include "edsel.h"
#include "error.h"
#include "memory.h"
#include "quest.h"
#include "sel.h"
#include "fe/message.h"
#include "fe/int.h"


typedef struct cblist_s
{
	struct cblist_s *Next;
	cbrush_t b;
} cblist_t;


static cblist_t *new_brush;


static int SubtractBrushes(cblist_t *b1,cbrush_t *b2)
{
	int i;
	cblist_t *nb;

	if (!CSG_BrushesIntersect(&b1->b,b2))
	{
		b1->Next=new_brush;
		new_brush=b1;
		return 0;
	}

	nb=Q_malloc(sizeof(cblist_t));
	if (!nb) return -2;
	memset(nb,0,sizeof(cblist_t));
	for (i=0;i<b2->num_faces;i++)
	{
		switch (CSG_SplitBrushWithFace(&b1->b,b2->faces[i],&nb->b))
		{
		case -3:
			return -2;

		case 0:
			b1->Next=new_brush;
			new_brush=b1;
			b1=nb;
			nb=Q_malloc(sizeof(cblist_t));
			if (!nb) return -2;
			memset(nb,0,sizeof(cblist_t));
			break;

		case -1: /* shouldn't happen, but won't hurt */
			b1->Next=new_brush;
			new_brush=b1;
			return 0;

		case -2:
			break;
		}
	}
	Q_free(nb);
	return 0;
}


static int DoSubtract(sel_t *sfrom,sel_t *swith)
{
	bsel_t *bs;
	cblist_t *with,*w;
	cblist_t *cb,*cn;
	cblist_t *from,*f,*fn;
	brush_t *b;

	with=NULL;
	for (bs=swith->bsel;bs;bs=bs->Next)
	{
		cb=Q_malloc(sizeof(cblist_t));
		if (!cb) return -1;
		memset(cb,0,sizeof(cblist_t));
		if (!CSG_BrushToCBrush(&cb->b,bs->Brush))
			return -1;
		cb->Next=with;
		with=cb;
	}

	for (bs=sfrom->bsel;bs;bs=bs->Next)
	{
		from=Q_malloc(sizeof(cblist_t));
		if (!from) return -1;
		memset(from,0,sizeof(cblist_t));
		if (!CSG_BrushToCBrush(&from->b,bs->Brush))
			return -1;

		for (w=with;w;w=w->Next)
		{
			new_brush=NULL;
			for (f=from;f;f=fn)
			{
				fn=f->Next;
				if (SubtractBrushes(f,&w->b)==-2)
					return -1;
			}
			from=new_brush;
		}

		for (;from;from=fn)
		{
			fn=from->Next;

			b=CSG_BuildBrush(NULL,&from->b);
			if (!b) return -1;
			b->Entity=bs->Brush->Entity;
			b->Group=bs->Brush->Group;
			B_Link(b);
			Q_free(from);
		}
		
		B_Unlink(bs->Brush);
		B_Free(bs->Brush);
	}

	for (cb=with;cb;cb=cn)
	{
		cn=cb->Next;
		CSG_FreeBrush(&cb->b);
		Q_free(cb);
	}
	return 0;
}


static void C_subtract(void)
{
	sel_t Smashlist;
	bsel_t *b1,*b2;
	int i;

	if (M.sel.num_bselected==0)
	{
		S_NewMessage("CSG Subtract: No brushes selected!");
		return;
	}

	for (b1=M.sel.bsel;b1;b1=b1->Next)
		if (b1->Brush->bt->type!=BRUSH_NORMAL)
		{
			S_NewMessage("CSG Subtract: Can only use normal brushes for CSG operations!");
			return;
		}

	Smashlist=M.sel;
	memset(&M.sel,0,sizeof(sel_t));
	M.sel.flags=SELECTION_PRIMARY;
	Sel_Update(&M.sel);

	i=S_I_GetSelection("Select the subtractor(s).");

	if (!i || M.sel.num_bselected==0)
	{
		ClearSelAll();
		S_NewMessage("CSG Subtract: No subtractors, command aborted.");
		M.sel = Smashlist;
		return;
	}
	for (b1=M.sel.bsel;b1;b1=b1->Next)
	{
		if (b1->Brush->bt->type!=BRUSH_NORMAL)
		{
			S_NewMessage("CSG Subtract: Can only use normal brushes for CSG operations!");
			Sel_Clear(&Smashlist);
			return;
		}
		
		for (b2=Smashlist.bsel;b2;b2=b2->Next)
		{
			if (b2->Brush==b1->Brush)
			{
				Error_Add("Cannot have the same brush in 'from' and 'with' lists!");
				Sel_Clear(&Smashlist);
				return;
			}
		}
	}

	DoSubtract(&Smashlist,&M.sel);

	Sel_Clear(&Smashlist);

	S_NewMessage("Subtraction performed.");
}


static func_t F_csg_subtract=
{"csg.subtract",FUNC_VOID,{fv:C_subtract},NULL,NULL,0,
{"CSG subtract.",
"This command does a CSG subtraction. This means that a set of brushes "
"will be removed from another set of brushes, i.e. the second set of "
"brushes will be changed so they no longer overlap the first set.\n"
"\n"
"Before using it, select the brushes you want to subtract from. You will "
"then be asked to select the brushes you want to subtract with. After the "
"subtraction has been done, the second set of brushes will be selected.\n"
"\n"
"This command can be very useful, but it's also dangerous. Round-off "
"errors may cause leaks and other problems in your map (if you use Quest's "
"new .map format you should be safe, though), so you'll probably want to be "
"careful with it and snap the new brushes to the grid after subtracting from "
"them."}
};

static func_t *funcs[]={&F_csg_subtract};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"csg.subtract","CSG subtract command."}};

PLUGIN_FOOT_FUNCS

