#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "camera.h"
#include "config.h"
#include "error.h"
#include "quest.h"
#include "status.h"
#include "transform.h"
#include "undo.h"
#include "fe/3d.h"
#include "fe/int.h"
#include "fe/message.h"


static void C_edit_move(qatom_t l)
{
	vec3_t delta;
	int dir=QINT(QLIST(&l,0));

	if (!M.sel.vsel && !M.sel.esel && !M.sel.bsel && !M.sel.fsel)
	{
		S_NewMessage("edit.move: Nothing selected!");
		return;
	}

	Move90(S_cur_vport,dir,&delta,cv_edit_snap_size->i);

	TransformMove(delta,&M.sel);
}
static func_t F_edit_move=
{"edit.move",FUNC_VOID,{fva:C_edit_move},NULL,"i",0,
{"Moves the current selection.",
"This command moves the current selection in some "
"direction. The distance moved is controlled by edit.snap_size."
"\n\n"
"If any vertices are selected, they will be moved. If no vertices are "
"selected, all vertices of all selected brushes and faces are moved. "
"If any entities are selected, they will also be moved. If they have "
"an origin field, it will be updated. If an entity has any brushes attached, "
"they will also be moved. If a brush is part of a selected entity, and is "
"selected itself, the entire brush will be moved even if only some of its "
"vertices are selected."}
};


/*
TODO:
how much should we move? should diffent grid sizes require different amounts
of mouse movement to move one grid-multiple. zooming? 3d? configurable?
*/
static void C_edit_move_drag(void)
{
	s_event_t e;
	int tx,ty;
	int sx,sy,ox,oy;
	vec3_t left,up,delta;

	if (!M.sel.vsel && !M.sel.esel && !M.sel.bsel && !M.sel.fsel)
		return;

	Move90(S_cur_vport,MOVE_LEFT,&left,cv_edit_snap_size->i);
	Move90(S_cur_vport,MOVE_UP,&up,cv_edit_snap_size->i);

	S_I_Setup(&e,S_I_MOTION|S_I_RELEASE,NULL);
	tx=ty=0;
	ox=oy=0;
	while (1)
	{
		S_I_Wait(&e);
		if (e.type==S_I_CLICK || e.type==S_I_RELEASE)
			break;
		if (e.type==S_I_MOTION)
		{
			tx+=e.dx;
			ty+=e.dy;

			sx=-tx/4-ox;
			sy=-ty/4-oy;
			delta.x=sx*left.x+sy*up.x;
			delta.y=sx*left.y+sy*up.y;
			delta.z=sx*left.z+sy*up.z;
			ox+=sx;
			oy+=sy;
			TransformMove(delta,&M.sel);

			Undo_Update();
			S_UpdateAllViewports();
		}
	}
	S_I_Done();
}
static func_t F_edit_move_drag=
{"edit.move_drag",FUNC_VOID,{fv:C_edit_move_drag},NULL,NULL,0,
{"Moves the current selection by tracking the mouse."} /* TODO */
};


static func_t *funcs[]={&F_edit_move,&F_edit_move_drag};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"edit.move","Fundamental selection movement commands."}};

PLUGIN_FOOT_FUNCS

