#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "config.h"
#include "edsel.h"
#include "error.h"
#include "quest.h"
#include "status.h"
#include "fe/popup.h"
#include "fe/render.h"


static void UpdateRFlags(int change,int add,int del)
{
	int oldf;
	int rem;

	oldf=cv_render_flags->i;
	if (!(cv_render_flags->i&change))
	{
		Cfg_SetInt(cv_render_flags,cv_render_flags->i|change|add);
	}
	else
	{
		rem=change|del;
		Cfg_SetInt(cv_render_flags,cv_render_flags->i&(~rem));
		rem&=oldf;
		if (rem&RENDER_ENTITY_FOCAL) ClearSelEnts();
		if (rem&RENDER_FACE_VERT)
		{
			ClearSelVerts();
			ClearSelFaces();
		}
		if (rem&RENDER_BRUSH_VERT)
		{
			ClearSelVerts();
			ClearSelBrushes();
		}
	}
	S_R_SetUpdate(R_UPDATE_ALL);
}

typedef struct
{
	unsigned int flag;
	const char *str;
	unsigned int add,del;
} render_flag_t;

static render_flag_t rflags[]={
{},
{RENDER_BRUSH,"Brushes",
	0,
	RENDER_BRUSH_FOCAL|RENDER_BRUSH_VERT},
{RENDER_BRUSH_FOCAL,"Brush focals",
	RENDER_BRUSH,
	RENDER_BRUSH_VERT},
{RENDER_BRUSH_VERT,"Brush vertices",
	RENDER_BRUSH|RENDER_BRUSH_FOCAL,
	0},

{RENDER_FACE_FOCAL,"Face focals",
	RENDER_BRUSH,
	RENDER_FACE_VERT},
{RENDER_FACE_VERT,"Face vertices",
	RENDER_BRUSH|RENDER_FACE_FOCAL,
	0},
{},
{},

{RENDER_ENTITY,"Entities",
	0,
	RENDER_ENTITY_FOCAL|RENDER_ENTITY_LINK},
{RENDER_ENTITY_FOCAL,"Entity focals",
	RENDER_ENTITY,
	RENDER_ENTITY_LINK},
{RENDER_ENTITY_LINK,"Entity links",
	RENDER_ENTITY|RENDER_ENTITY_FOCAL,
	0}
};
#define N_RFLAGS ((signed)(sizeof(rflags)/sizeof(rflags[0])))

static void C_render_flags_toggle(qatom_t l)
{
	int i=QINT(QLIST(&l,0));
	if (i<0 || i>=N_RFLAGS || !rflags[i].flag)
		ERROR("Invalid flag %i",i);
	UpdateRFlags(rflags[i].flag,rflags[i].add,rflags[i].del);
}
static func_t F_render_flags_toggle=
{"render.flags.toggle",FUNC_VOID,{fva:C_render_flags_toggle},NULL,"i",0};


static void C_render_flags_change(void)
{
/* using popup's here is a bit of a hack, but it should work quite nicely */
	int i,j;
	S_Popup_Init();
	for (i=0;i<N_RFLAGS;i++)
		if (rflags[i].str)
			S_Popup_AddStr("%s\t%c",
				rflags[i].str,
				cv_render_flags->i&rflags[i].flag?'1':'0');
	j=S_Popup_Display();
	S_Popup_Free();
	if (j==-1) return;
	for (i=0;i<N_RFLAGS;i++)
		if (rflags[i].str)
		{
			if (!j)
			{
				UpdateRFlags(rflags[i].flag,rflags[i].add,rflags[i].del);
				break;
			}
			else
				j--;
		}
}
static func_t F_render_flags_change=
{"render.flags.change",FUNC_VOID,{fv:C_render_flags_change},NULL,NULL,0};


static func_t *funcs[]={&F_render_flags_toggle,&F_render_flags_change};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"render.flags.cmds","Commands to change the rendering flags."}};

static int Init(void)
{
	Func_Register(&plugin,funcs,sizeof(funcs));
	UpdateRFlags(0,0,0);
	return 1;
}

PLUGIN_FOOT

