#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "quest/defines.h"
#include "quest/types.h"

#include "quest/fe/3d.h"
#include "quest/fe/gui.h"
#include "quest/fe/input.h"
#include "quest/fe/main.h"
#include "quest/fe/message.h"
#include "quest/fe/quest.h"
#include "quest/fe/render.h"
#include "quest/version.h"

#include "../signals.h"


static int should_exit=0;

void S_Exit(void)
{
	should_exit=1;
}

void S_SetMapName(const char *name)
{
	(void)name;
//	printf("S_SetMapName('%s')\n",name);
}


void S_NewMessage(const char *format, ...)
{
	va_list args;
	printf("| ");
	va_start(args,format);
	vprintf(format,args);
	va_end(args);
	printf("\n");
}


#include "quest/brush.h"

int /* TODO */
	S_R_size_brush=sizeof(brush_t),
	S_R_size_face=sizeof(face_t),
	S_R_size_entity=sizeof(entity_t);

void S_R_EntityUpdate(entity_t *e) { (void)e; }
void S_R_EntityFree(entity_t *e) { (void)e; }
void S_R_BrushUpdate(struct brush_s *b) { (void)b; }
void S_R_BrushFree(struct brush_s *b) { (void)b; }
void S_R_FaceUpdate(struct face_s *f) { (void)f; }
void S_R_FaceFree(struct face_s *f) { (void)f; }
void S_R_SetUpdate(unsigned int flags) { (void)flags; }
void S_R_Map_Free(void) {}


int S_GUI_Dialog(const char *title,const char *text, ...)
{
	printf("S_GUI_Dialog() %s: %s",title,text);
	abort(); /* TODO */
}
void S_GUI_DialogOK(const char *title,const char *format, ...)
{
	va_list args;
	printf("S_GUI_DialogOK() %s: ",title);
	va_start(args,format);
	vprintf(format,args);
	va_end(args);
	printf("\n");
}
int S_GUI_YesNo(const char *title,const char *format, ...)
{
	va_list args;
	printf("S_GUI_DialogYesNo() %s: ",title);
	va_start(args,format);
	vprintf(format,args);
	va_end(args);
	printf("\n");
	abort();
}
int S_GUI_FileDialog(const char *title,const char *button,
                     const char *wildcard,char *name)
{
	printf("S_GUI_FileDialog('%s','%s','%s','%s')\n",
		title,button,wildcard,name);
	fgets(name,1024 /* TODO */,stdin);
	while (name[0] && name[strlen(name)-1]<=32) name[strlen(name)-1]=0;
	if (name[0]) return 1;
	return 0;
}
int S_GUI_EditEntity(struct entity_s *e)
{
	printf("S_GUI_EditEntity: %p\n",e);
	abort();
}
int S_GUI_PopEntity(const char *title,char **keys,char **values,int num_keys)
{
	int i;
	printf("S_GUI_PopEntity('%s')\n",title);
	for (i=0;i<num_keys;i++)
	{
		printf("   %10s %10s\n",keys[i],values[i]);
	}
	abort();
}


static viewport_t viewport;
viewport_t *S_cur_vport=&viewport;

void S_UpdateCurViewport(void) {}
void S_UpdateAllViewports(void) {}
void S_R_ForEachViewport(void (*func)(struct viewport_s *)) { func(&viewport); }

#if 0
struct S_Menu *S_RootMenu=(struct S_Menu *)1;
struct S_Menu *S_Menu_Submenu(struct S_Menu *m,const char *name)
{
	printf("S_Menu_Sebmenu(%p,'%s')\n",m,name);
	return (struct S_Menu *)1;
}
void S_Menu_AddItem(struct S_Menu *m,const char *name,const char *cmd,int sep)
{
	printf("S_Menu_AddItem(%p,'%s',%p,%i)\n",m,name,cmd,sep);
}
void S_Key_Bind(const char *key,const char *cmd)
{
	printf("S_Key_Bind('%s',%p)\n",key,cmd);
}
#endif
#include "quest/qfunc.h"
#include "quest/qsymbol.h"
static void C_menu_add(qatom_t *ret,qatom_t a)
{
//	printf("menu.add: '%s' to %i:%p\n",a.x.l[1].x.string,a.x.l[0].type,a.x.l[0].x.ptr);
	ret->type=QAT_MENU;
	ret->x.ptr=(void *)0x1234;
	(void)a;
}
static func_t F_menu_add=
{"menu.add",FUNC_ATOM,{faa:C_menu_add},NULL,"[nm]c",FUNC_NOT_USER};

static void C_menu_set(qatom_t a)
{
//	printf("menu.set: %i:%p\n",a.x.l[0].type,a.x.l[0].x.ptr);
	(void)a;
}
static func_t F_menu_set=
{"menu.set",FUNC_VOID,{fva:C_menu_set},NULL,"[nm]",FUNC_NOT_USER};

static void C_menu(qatom_t *ret,qatom_t a)
{
//	printf("menu: %i:%s\n",a.x.l[0].type,a.x.l[0].type==QAT_STRING?a.x.l[0].x.string:QL_SymbolName(a.x.l[0].x.symbol));
	ret->type=QAT_MENU;
	ret->x.ptr=(void *)0x2345;
	(void)a;
}
static func_t F_menu=
{"menu",FUNC_ATOM,{faa:C_menu},NULL,"[nc]",FUNC_NOT_USER};

static void C_bind(qatom_t a)
{
	func_t *f;
	f=a.x.l[0].x.func;
/*	int i;
	printf("bind: %p ('%s')\n",f,QL_SymbolName(f->symbol));
	for (i=1;i<a.type;i++)
	{
		QA_Describe(&a.x.l[i],4);
	}*/
}
static func_t F_bind=
{"bind",FUNC_VOID,{fva:C_bind},NULL,"f[mc]+",FUNC_NOT_USER};

static func_t *funcs[]={&F_menu_add,&F_menu_set,&F_menu,&F_bind};


int main(int argc, char **argv)
{
	printf("Quest (text) starting\n");
   S_InitSignals(); /* TODO? */
   InitQuestCore();
	Func_Register(NULL,funcs,sizeof(funcs));

   S_NewMessage("Quest (text) initialized (%s).",QUEST_VERSION);

   InitQuest(argc,argv);
	/* TODO */
	{ char buf[256]; fgets(buf,sizeof(buf),stdin); }
   DoneQuest();

   return 0;
}

/*

ql system
 hooks/ql functions
 cvars/symbols
?
frontend
?
heaps?
plugin system
initialize map
	viewports
	undo system
load config 'Quest'
final command line parsing (load map)

*/

