#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "quest/defines.h"
#include "quest/types.h"

#include "quest/fe/texture.h"

#include "quest/memory.h"


typedef struct
{
	texture_t tex;
	int load;
} txt_texture_t;


static tex_funcs_t *tf;


static txt_texture_t **tex;
static int num_tex;

int S_tex_use_cache;


static txt_texture_t *FindTexture(const char *name)
{
   int l,h,m;
   int i;

   if (!num_tex) return NULL;

   for (l=0,h=num_tex-1;l<h;)
   {
      m=(l+h)/2;
      if (!(i=stricmp(name,tex[m]->tex.name)))
         return tex[m];

      if (i<0) h=m-1; else l=m+1;
   }
   if (!stricmp(name,tex[l]->tex.name))
      return tex[l];
   return NULL;
}


static int S_Txt_WantLoad(txt_texture_t *t,int level)
{
   if (level<=t->load) return 1;

   if (t->load==-1)
      return 0;

   if (!tf)
      return 0;

   if (level==1 && tf->load_size)
   {
      tf->load_size(t,&t->tex);
      if (t->load==1)
         return 1;
   }

   if (!tf->load)
      return 0;

   tf->load(t,&t->tex);
   if (t->load<level)
   {
      t->load=-1;
      return 0;
   }
   return 1;
}


static txt_texture_t *InsertTexture(const char *name,txt_texture_t *t)
{
   int l,h,m;
   int i;

   txt_texture_t **tt;

   if (num_tex)
   {
      for (l=0,h=num_tex-1;l<h;)
      {
         m=(l+h)/2;
         if (!(i=stricmp(name,tex[m]->tex.name)))
            return NULL;

         if (i<0) h=m-1; else l=m+1;
      }
      i=stricmp(name,tex[l]->tex.name);
      if (!i) return 0;
      if (i>0) l++;
   }
   else
      l=0;

   tt=Q_realloc(tex,sizeof(txt_texture_t *)*(num_tex+1));
   if (!tt) return NULL;
   tex=tt;

   for (h=num_tex;h>l;h--)
      tex[h]=tex[h-1];
   tex[h]=t;
   num_tex++;

   return t;
}


texture_t *S_T_Load(const char *name)
{
   txt_texture_t *t;

   t=FindTexture(name);
   if (!t)
      return NULL;
   t->tex.cached=1;
   if (!S_Txt_WantLoad(t,2))
      return NULL;
   return &t->tex;
}

texture_t *S_T_LoadSize(const char *name)
{
   txt_texture_t *t;

   t=FindTexture(name);
   if (!t)
      return NULL;
   t->tex.cached=1;
   if (!S_Txt_WantLoad(t,1))
      return NULL;
   return &t->tex;
}

int S_T_LoadCheck(const char *name)
{
   if (FindTexture(name))
      return 1;
   return 0;
}


void S_T_GetDescription(char *dst,const texture_t *t)
{
   if (!tf->tex_desc)
   {
      *dst=0;
      return;
   }
   tf->tex_desc(dst,t);
}


int S_T_Cache(const char *name)
{
   txt_texture_t *t;
   if (!S_tex_use_cache) return 1;
   t=FindTexture(name);
   if (!t)
      return 0;
   t->tex.cached=1;
   return 1;
}

int S_T_Uncache(const char *name)
{
   txt_texture_t *t;
   if (!S_tex_use_cache) return 1;
   t=FindTexture(name);
   if (!t)
      return 0;
   t->tex.cached=0;
   return 1;
}


void S_T_Init(int cache,tex_funcs_t *ptf)
{
   S_tex_use_cache=cache;
   S_T_Clear();
   tf=ptf;
}

void S_T_Clear(void)
{
   int i;
   tf=NULL;
   for (i=0;i<num_tex;i++)
   {
      if (tex[i]->tex.g)
         Q_free(tex[i]->tex.g);
      Q_free(tex[i]);
   }
   Q_free(tex);
   tex=NULL;
   num_tex=0;
}


void S_T_ClearCache(void)
{
   int i;
   if (!S_tex_use_cache) return;
   for (i=0;i<num_tex;i++)
      tex[i]->tex.cached=0;
}


void S_T_SetPalette(unsigned char *pal)
{
	(void)pal;
}


texture_t *S_T_GetTexture(void *tex)
{
   txt_texture_t *t=tex;
   return &t->tex;
}


void *S_T_AddTexture(const char *name,int width,int height,void *g)
{
   txt_texture_t *t;

   if (FindTexture(name))
      return 0;

   t=Q_malloc(sizeof(txt_texture_t));
   if (!t)
      return 0;
   memset(t,0,sizeof(txt_texture_t));
   strcpy(t->tex.name,name);
   t->tex.width=width;
   t->tex.height=height;
   t->tex.g=g;

   if (!S_tex_use_cache)
      t->tex.cached=1;

   if (width==-1)
      t->load=0;
   else
      t->load=1;

   return InsertTexture(name,t);
}


int S_T_AddSize(void *tex,int width,int height)
{
   txt_texture_t *t=tex;

   if (t->load>=1) return 0;
   if (width==-1) return 0;
   t->tex.width=width;
   t->tex.height=height;
   t->load=1;
   return 1;
}

int S_T_AddData8(void *tex,int width,int height,
	ARG_UNUSED(int data_x),ARG_UNUSED(int data_y),
	ARG_UNUSED(unsigned char *data),
	ARG_UNUSED(unsigned char *ppal))
{
   return S_T_AddData24(tex,width,height,0,0,0);
}

int S_T_AddData24(void *tex,int width,int height,
	ARG_UNUSED(int data_x),ARG_UNUSED(int data_y),
	ARG_UNUSED(unsigned char *data))
{
   txt_texture_t *t=tex;

   t->tex.width=width;
   t->tex.height=height;

   t->load=2;

   return 1;
}

