#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <setjmp.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#if OLD_CMD_STUFF

#include "error.h"
#include "hook.h"
#include "undo.h"
#include "fe/3d.h"
#include "fe/gui.h"
#include "fe/input.h"
#include "fe/menu.h"


/*
TODO: use cvars to control keyboard and menu layout. Treat the values to
Cmd_Register as defaults when creating the cvars.
*/

static cmd_reg_t *cmds=NULL;
static int num_cmds=0;


cmd_reg_t *Cmd_Lookup(const char *name)
{
   cmd_reg_t *c;
   int i;

   if (!name) return NULL;
   for (i=0,c=cmds;i<num_cmds;i++,c++)
      if (!strcmp(name,c->name))
         return c;
   return NULL;
}


static int in_command;
static int cur_filter;
static jmp_buf cmd_jmp_buf;

void Cmd_Run(const char *name)
{
   cmd_reg_t *c;

	/* TODO - !!HACK!! */
	QL_RunCmd(name);
	return;

#if 0
   c=Cmd_Lookup(name);
   if (!c)
   {
      S_Error("Cmd_Run","No such command: '%s'",name);
      return;
   }

   if (!in_command)
   {
      if (setjmp(cmd_jmp_buf))
      {
         Undo_Abort();
         S_UpdateAllViewports();
         in_command=0;
         return;
      }

      if (Hook_Call(hk_cmd_start,c))
         return;
         
      in_command++;
      if (!(c->f.flags&(CMD_NOUNDO|CMD_UNDO_CLEAR)))
      {
         if (c->f.undo_name)
            Undo_New(c->f.undo_name,1,0);
         else
            Undo_New(name,0,0);
      }
      else
      if (c->f.flags&CMD_NOUNDO)
      {
         Undo_New(NULL,0,1);
      }
      else
      {
         Undo_New(NULL,0,2);
      }
      Hook_Call(c->hook,NULL);
      Hook_Call(hk_cmd_end,c);
      Undo_Done();
      Cmd_HandleRedraw();
      in_command--;
      return;
   }

   in_command++;
   if (cur_filter && !(c->f.flags&cur_filter))
      return;
   if (c->f.flags&(CMD_NOUNDO|CMD_UNDO_CLEAR))
      return;
   Hook_Call(c->hook,NULL);
   Cmd_HandleRedraw();
   in_command--;
#endif
}


void Cmd_Push(int filter)
{
   ASSERT(in_command==1);
   cur_filter=filter;
}

void Cmd_Pop(void)
{
   ASSERT(in_command==1);
   cur_filter=0;
}


void Cmd_Register(struct plugin_s *p,cmd_reg_t *cmd,int size)
{
   int i;
   char buf[128];
   
   size/=sizeof(cmd_reg_t);
   cmds=realloc(cmds,sizeof(cmd_reg_t)*(num_cmds+size));
   if (!cmds) ABORT("Cmd_Register: Out of memory!");
   memcpy(&cmds[num_cmds],cmd,size*sizeof(cmd_reg_t));
   num_cmds+=size;

   cmd=&cmds[num_cmds-size];
   for (i=0;i<size;i++,cmd++)
   {
/*      {
         printf("%-30s %-20s %-40s : %s\n",
            cmd->name,cmd->d.key?cmd->d.key:"",
            cmd->d.menu?cmd->d.menu:"",
            cmd->desc?cmd->desc:"");
      }*/
#if 0
      if (cmd->d.key)
         S_Key_Bind(cmd->d.key,cmd->name);
      if (cmd->d.menu)
      {
         char buf[128];
         const char *c;
         char *d;
         struct S_Menu *m;

         m=S_RootMenu;
         for (c=cmd->d.menu;*c;)
         {
            for (d=buf;*c && *c!='/';) *d++=*c++;
            *d=0;
            if (*c=='/')
            {
               m=S_Menu_Submenu(m,buf);
               c++;
            }
         }
         S_Menu_AddItem(m,buf,cmd->name,0);
      }
#endif

      snprintf(buf,sizeof(buf),"c.%s",cmd->name);
      cmd->hook=Hook_Define(buf,HOOK_CALL_HANDLE);
      Hook_AddFuncLast(cmd->hook,
         HOOK_FUNC_VOID,(hook_func_void_t)cmd->func,NULL,
         p);
   }
}


/* TODO: move command helper functions to different file? */

static int redraw;

void Cmd_SetRedraw(int i)
{
   if (i>redraw) redraw=i;
}

void Cmd_HandleRedraw(void)
{
   if (redraw==1)
      S_UpdateCurViewport();
   if (redraw==2)
      S_UpdateAllViewports();

   redraw=0;
}


#if 0
void Cmd_Abort(const char *proc,const char *format, ...)
{
   char buf[256];
   va_list args;

   va_start(args,format);
   vsnprintf(buf,sizeof(buf)-1,format,args);
   va_end(args);

   if (!in_command)
      S_Abort("Cmd_Abort","%s: %s",proc,buf);

   S_GUI_DialogOK("Cmd_Abort","%s: %s",proc,buf);
   longjmp(cmd_jmp_buf,1);
}
#endif
#endif

