#ifndef ERROR_H
#define ERROR_H

void Abort_Str(const char *str) __attribute__ ((noreturn));
void Error_Str(const char *str) __attribute__ ((noreturn));
void Error_Add_Str(const char *str);
void Warning_Str(const char *str);

void Abort(const char *format, ...)
	__attribute__ ((format(printf,1,2),noreturn));
void Error(const char *format, ...)
	__attribute__ ((format(printf,1,2),noreturn));
void Error_Add(const char *format, ...)
	__attribute__ ((format(printf,1,2)));
void Warning(const char *format, ...)
	__attribute__ ((format(printf,1,2)));

void Abort_File(const char *file,const char *function,int line,const char *format, ...)
	__attribute__ ((format(printf,4,5),noreturn));
void Error_File(const char *file,const char *function,int line,const char *format, ...)
	__attribute__ ((format(printf,4,5),noreturn));
void Error_Add_File(const char *file,const char *function,int line,const char *format, ...)
	__attribute__ ((format(printf,4,5)));
void Warning_File(const char *file,const char *function,int line,const char *format, ...)
	__attribute__ ((format(printf,4,5)));

#define ABORT(x, y...) \
	do { Abort_File(__FILE__,__FUNCTION__,__LINE__,x , ##y); } while (0)
#define ERROR(x, y...) \
	do { Error_File(__FILE__,__FUNCTION__,__LINE__,x , ##y); } while (0)
#define ERROR_ADD(x, y...) \
	do { Error_Add_File(__FILE__,__FUNCTION__,__LINE__,x , ##y); } while (0)
#define WARNING(x, y...) \
	do { Warning_File(__FILE__,__FUNCTION__,__LINE__,x , ##y); } while (0)

#define ABORT_MEM(x) \
	do { if (!(x)) ABORT("Out of memory!"); } while (0)
#define ERR_MEM(x) \
	do { if (!(x)) ERROR("Out of memory!"); } while (0)
#define ERR_AMEM() \
	do { ERROR_ADD("Out of memory!"); } while (0)

#ifdef ASSERT_DUMP_CORE

#define ASSERT(x)                                          \
	do { if (!(x))                                          \
	{                                                       \
	fprintf(stderr,"-- %s: Assert: '%s' (%s, line %i)",     \
		__FUNCTION__,#x,__FILE__,__LINE__);                  \
		fflush(stderr);                                      \
		(*(int *)-1)=0;                                      \
		S_Abort(__FUNCTION__,"Assert: '%s' (%s, line %i)",   \
			#x,__FILE__,__LINE__);                            \
	} } while (0)

#else

#define ASSERT(x) \
	do { if (!(x)) \
		Abort_File(__FILE__,__FUNCTION__,__LINE__,"Assert failed: %s",#x); \
	} while (0)

#endif


/* Displays all accumulated errors and clears the error list */
void Error_Display(const char *format, ...) __attribute__ ((format(printf,1,2)));

/* Clear the error list */
void Error_Clear(void);

#endif

