#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "game.h"
#include "game_reg.h"

#include "error.h"


static game_reg_t *G;


int G_LoadMap(const char *fname)
{
   if (!G)
   {
      Error_Add("G_LoadMap: No game active!");
      return 0;
   }
   if (!G->map.load)
   {
      Error_Add("G_LoadMap: The active game doesn't provide a map loader!");
      return 0;
   }
   return G->map.load(fname);
}

int G_SaveMap(const char *fname)
{
   if (!G)
   {
      Error_Add("G_SaveMap: No game active!");
      return 0;
   }
   if (!G->map.save)
   {
      Error_Add("G_SaveMap: The active game doesn't provide a map saver!");
      return 0;
   }
   return G->map.save(fname);
}


int G_Texdef_Init(struct texdef_s *t)
{
   if (G && G->texdef.init) return G->texdef.init(t);
   return 1;
}

void G_Texdef_Free(struct texdef_s *t)
{
   if (G && G->texdef.free) G->texdef.free(t);
}

int G_Texdef_Copy(struct texdef_s *dst,const struct texdef_s *src)
{
   if (G && G->texdef.copy) return G->texdef.copy(dst,src);
   return 1;
}


int G_Load(game_reg_t *g)
{
   if (G) return 0;
   G=g;
   return 1;
}

void G_Unload(void)
{
   G=NULL;
}

