#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "plugin.h"

#include "error.h"
#include "hook.h"


plugin_t **plugins;
int num_plugins;


void Plugin_Register(plugin_t *p)
{
   int i;
   int l,h,m;

   p->active=0;
   
   if (num_plugins)
   {
      for (l=0,h=num_plugins-1;l<h;)
      {
         m=(l+h)/2;
         if (!(i=strcmp(p->s.name,plugins[m]->s.name)))
            return;

         if (i<0) h=m-1; else l=m+1;
      }
      i=strcmp(p->s.name,plugins[l]->s.name);
      if (!i)
         return;
      if (i>0) l++;
   }
   else
      l=0;

   plugins=realloc(plugins,sizeof(plugin_t *)*(num_plugins+1));
   if (!plugins)
      ABORT("Plugin_Register: Out of memory!");
   for (h=num_plugins;h>l;h--)
      plugins[h]=plugins[h-1];
   plugins[h]=p;
   num_plugins++;
}


plugin_t *Plugin_Lookup(const char *name)
{
   int i;
   int l,h,m;

   if (!num_plugins)
      return NULL;
      
   for (l=0,h=num_plugins-1;l<h;)
   {
      m=(l+h)/2;
      if (!(i=strcmp(name,plugins[m]->s.name)))
         return plugins[m];

      if (i<0) h=m-1; else l=m+1;
   }
   i=strcmp(name,plugins[l]->s.name);
   if (!i)
      return plugins[l];
   return NULL;
}


void Plugin_Activate(plugin_t *p,struct config_s *c)
{
	ASSERT(p);
   if (p->active) return;
	p->cfg=c;
   if (p->f.init)
   {
      if (!p->f.init())
         return;
   }
   p->active=1;
}

static void Plugin_DeactivateOne(plugin_t *p)
{
	printf("unloading plugin '%s'\n",p->s.name);
   if (!p->active) return;
   if (p->f.done)
   {
      if (!p->f.done())
         return;
   }
   Hook_Call(hk_plugin_remove,p);
   p->active=0;
	p->cfg=NULL;
}

void Plugin_Deactivate(struct config_s *cfg)
{
	int i;
	for (i=0;i<num_plugins;i++)
	{
		if (!plugins[i]->active) continue;
		if (plugins[i]->cfg!=cfg) continue;
		Plugin_DeactivateOne(plugins[i]);
	}
}


void Plugin_Init(void)
{
/* TODO  use configuration settings to do this */
//   Plugin_ActivateBasic();
//   Plugin_Activate(Plugin_Lookup("games/quake"));
}

